import datanway as dataset
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Flatten, Input, Lambda
from tensorflow.keras.layers import Conv2D, MaxPooling2D
from tensorflow.keras.layers import BatchNormalization
from matchnn import MatchCosine
import tensorflow as tf

bsize = 32 # batch size
classes_per_set = 5 # classes per set or 5-way
samples_per_class = 1 # samples per class 1-shot


data = dataset.OmniglotNShotDataset(batch_size=bsize,classes_per_set=classes_per_set,samples_per_class=samples_per_class,trainsize=32000,valsize=10000)

# Image embedding using Deep Convolutional Network
def ConvEmbedding():
    #define the input shape
    x_in=Input(shape=(28,28,1),name='input')

    #Define the convolutional network
    x = Conv2D(64, (3, 3), padding='same', activation='relu',name='conv1')(x_in)
    x= BatchNormalization(name='bnorm1')(x)
    x= MaxPooling2D((2, 2), padding='same', name='mpool1')(x)
    x= Conv2D(64, (3, 3), padding='same', activation='relu',name='conv2')(x)
    x = BatchNormalization(name='bnorm2')(x)
    x= MaxPooling2D((2, 2), padding='same',name='mpool2')(x)
    x= Conv2D(64, (3, 3), padding='same', activation='relu',name='conv3')(x)
    x = BatchNormalization(name='bnorm3')(x)
    x = MaxPooling2D((2, 2), padding='same',name='mpool3')(x)
    x= Conv2D(64, (3, 3), padding='same', activation='relu',name='conv4')(x)
    x = BatchNormalization(name='bnorm4')(x)
    x = MaxPooling2D((2, 2), padding='same', name='mpool4')(x)
    fltn = Flatten(name='flatten')(x)

    #define the CNN model
    op_model =Model(inputs=x_in, outputs=fltn)

    return op_model

def slice_(x, index=0):
    return x[:,index,:,:,:]


numsupportset = samples_per_class*classes_per_set  # Total number of samples per support set
input1 = Input((numsupportset+1,28,28,1)) # Overall input shape to the CNN embedding model


conv_embedding=ConvEmbedding() # CNN embedding model

# Model definitions for  the embeddings for each input
modelinputs = []
for lidx in range(numsupportset):
    modelinputs.append(conv_embedding(Lambda(slice_, arguments={'index':lidx})(input1)))  #support set embeddings
targetembedding = conv_embedding(Lambda(lambda x: x[:,-1,:,:,:])(input1))          #target embeddings
modelinputs.append(targetembedding)   #Append both support set embeddings and target embedding to a single vector

supportlabels = Input((numsupportset,classes_per_set))  # Support set labels  (one hot encoded)
modelinputs.append(supportlabels)  # Append support set labels

knnsimilarity = MatchCosine(nway=classes_per_set)(modelinputs) # Model definition for the cosine distance based similarity

model = Model(inputs=[input1,supportlabels],outputs=knnsimilarity) # Overall model
model.compile(optimizer='adam',loss='sparse_categorical_crossentropy',metrics=['accuracy'], run_eagerly=True) # compile
model.run_eagerly=True

#Meta train set
print(data.datasets_cache["train"][0].shape) #Support set images (1:5) and Target set image (6th)
print(data.datasets_cache["train"][1].shape) #Support set labels (one hot encoded)
print(data.datasets_cache["train"][2].shape) #Target set labels (just the label number)

#Meta validation set
print(data.datasets_cache["val"][0].shape) #Support set images (1:5) and Target set image (6th)
print(data.datasets_cache["val"][1].shape) #Support set labels (one hot encoded)
print(data.datasets_cache["val"][2].shape) #Target set labels (just the label number)


#Fit the model
model.fit([data.datasets_cache["train"][0],data.datasets_cache["train"][1]],data.datasets_cache["train"][2],
          validation_data=[[data.datasets_cache["val"][0],data.datasets_cache["val"][1]],data.datasets_cache["val"][2]],
          epochs=10,batch_size=32,verbose=1)



